/* xdirname.c: return the directory part of a path.

Copyright (C) 1999 Free Software Foundation, Inc.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* Return directory for NAME.  This is "." if NAME contains no directory
   separators (should never happen for selfdir), else whatever precedes
   the final directory separator, but with multiple separators stripped.
   For example, `xdirname ("/foo//bar.baz")' returns "/foo".  Always
   return a new string.  */

#include <kpathsea/config.h>
#include <kpathsea/c-pathch.h>

#ifdef HAVE_GETURL
unsigned url_part_len P1C(const_string, name)
{
  int limit;
  for(limit = 0; name[limit] && name[limit] != ':'; limit++)
    ;
  if (name[limit] != ':'
      || !IS_DIR_SEP(name[limit+1])
      || !IS_DIR_SEP(name[limit+2])) {
    limit = 0;
  }
  else {
    for(limit +=3; name[limit] && !IS_DIR_SEP(name[limit]); limit++)
      ;
    if (!IS_DIR_SEP(name[limit]))
      limit = 0;
    else
      limit++;
  }

  return limit;
}
#endif
#ifdef WIN32
unsigned unc_part_len P1C(const_string, name)
{
  int limit;

  for (limit = 2; name[limit] && !IS_DIR_SEP(name[limit]); limit++)
    ;
  if (name[limit]) {
    for (limit++ ; name[limit] && !IS_DIR_SEP(name[limit]); limit++)
      ;
    if (!IS_DIR_SEP(name[limit]))
      limit = 0;
    else
      limit++;
  } else {
    /* malformed UNC name, backup */
    limit = 0;
  }
  return limit;
}
#endif

#ifdef WIN32
string
xdirsimplify P1C(string, name)
{
    char *begin, *p, *q, *end;
    
    if (*name == '\0') return name;
    end = name + strlen(name);
    begin = name;

    if (NAME_BEGINS_WITH_DEVICE(begin)) {
	begin += 2;
#ifdef HAVE_GETURL
    } else if (IS_URL_NAME(begin)) {
	begin += url_part_len(begin) - 1;
#endif
#if defined(WIN32) || defined(__CYGWIN__)
    } else if (IS_UNC_NAME(begin)) {
	begin += unc_part_len(begin) - 1;
    }
#endif

    /* Remove multiple path separators */
    p = begin;
    while (p != end) {
	if (IS_DIR_SEP(*p)) {
	    q = p + 1;
	    while (IS_DIR_SEP(*q)) q++;
	    if (q > p + 1)
		memmove(p + 1, q, end - q + 1);
	}
	p++;
    }
       
    /* Simplify path, removing /../ and /./ */
    p = begin;

    while (p != end) {
	if (IS_DIR_SEP(*p)) {
	    if (*(p+1) == '.') {
		if (IS_DIR_SEP(*(p+2))) {
		    memmove(p, p + 2, end - p + 1);
		    p = p - 1;
		}
		else if (*(p+2) == '.' && IS_DIR_SEP(*(p+3))) {
		    char *q;
		    if (p == begin) {
			/* Remove /../ at beginning */
			memmove(p, p+3, end - (p + 3) + 1);
			p--;
		    }
		    else {
			q = p - 1;
			while (q > begin && !IS_DIR_SEP(*q)) q--;
			if (IS_DIR_SEP(*q)) {
			    memmove(q, p + 3, end - (p + 3) + 1);
			    p = q - 1;
			}
		    }
		}
	    }
	}
	p++;
    }

    return name;
}
#endif /* WIN32 */

string
xdirname P1C(const_string, name)
{
  string ret;
  unsigned limit = 0, loc;

  /* Ignore a NULL name. */
  if (!name)
      return NULL;
  
  if (NAME_BEGINS_WITH_DEVICE(name)) {
      limit = 2;
#ifdef HAVE_GETURL
  } else if (IS_URL_NAME(name)) {
    limit = url_part_len(name);
#endif
#ifdef WIN32
  } else if (IS_UNC_NAME(name)) {
    limit = unc_part_len(name);
#endif
  }

  for (loc = strlen (name); loc > limit && !IS_DIR_SEP (name[loc-1]); loc--)
      ;

  if (loc == limit && limit > 0) {
      if (limit == 2) {
          ret = (string)xmalloc(limit + 2);
          ret[0] = name[0];
          ret[1] = name[1];
          ret[2] = '.';
          ret[3] = '\0';
      } else {
          ret = (string)xmalloc(limit + 2);
          strcpy(ret, name);
      }
  } else {
      /* If have ///a, must return /, so don't strip off everything.  */
      while (loc > limit+1 && IS_DIR_SEP (name[loc-1])) {
          loc--;
      }
      ret = (string)xmalloc(loc+1);
      strncpy(ret, name, loc);
      ret[loc] = '\0';
  }
    
  return ret;
}

#ifdef TEST

char *tab[] = {
  "\\\\neuromancer.ese-metz.fr\\fptex\\bin\\win32\\kpsewhich.exe",
  "\\\\neuromancer\\fptex\\win32\\kpsewhich.exe",
  "\\\\neuromancer\\fptex\\kpsewhich.exe",
  "\\\\neuromancer\\kpsewhich.exe",
  "p:\\bin\\win32\\kpsewhich.exe",
  "p:\\win32\\kpsewhich.exe",
  "p:\\kpsewhich.exe",
  "p:bin\\win32\\kpsewhich.exe",
  "p:win32\\kpsewhich.exe",
  "p:kpsewhich.exe",
  "p:///kpsewhich.exe",
  "/usr/bin/win32/kpsewhich.exe",
  "/usr/bin/kpsewhich.exe",
  "/usr/kpsewhich.exe",
  "///usr/kpsewhich.exe",
  "///kpsewhich.exe",
  "ftp://neuromancer.ese-metz.fr/fptex/bin/win32/kpsewhich.exe",
  "ftp://neuromancer/fptex/win32/kpsewhich.exe",
  "ftp://neuromancer/fptex/kpsewhich.exe",
  "ftp://neuromancer/kpsewhich.exe",
  NULL 
};

int main()
{
  char **p;
  for (p = tab; *p; p++)
    printf("name %s, dirname %s\n", *p, xdirname(*p));
  return 0;
}
#endif /* TEST */


/*
Local variables:
standalone-compile-command: "gcc -g -I. -I.. -DTEST variable.c kpathsea.a"
End:
*/

